/*
 * SignSample.java
 */

import java.io.*;
import java.util.*;
import java.security.*;
import java.security.cert.Certificate;
import javax.xml.crypto.*;
import javax.xml.crypto.dom.*;
import javax.xml.crypto.dsig.*;
import javax.xml.crypto.dsig.dom.*;
import javax.xml.crypto.dsig.spec.*;
import javax.xml.crypto.dsig.keyinfo.*;
import javax.xml.parsers.*;
import javax.xml.transform.*;
import javax.xml.transform.dom.*;
import javax.xml.transform.stream.*;

import org.w3c.dom.*;								// DocumentNXɗp
import org.apache.commons.codec.binary.Base64;		// Base64Envelopingŗp

public class SignSample {

    // ---------------------------------------------------------------------------
    // C֐.
	public static void main(String[] args) {

		// ^Cg\ƈmF
		System.out.println("Java XmlDsig Test");
		if(args.length < 4)
		{
			System.out.println("usage:");
			System.out.println("> java SignSample <target file> <pkcs12 file> <passwd> <output file>");
			return;
		}
		String targetFile  = args[0];
		System.out.println("target file : " + targetFile);
		String certFile   = args[1];
		String passwd     = args[2];	// \Ȃ
		System.out.println("sign cert   : " + certFile);
		String outputFile = args[3];
		System.out.println("output file : " + outputFile);

    	try
    	{
			// 
			boolean rc = SignTest(targetFile, certFile, passwd, outputFile);
			if(!rc)
			{
				System.out.println("Sign Error.");
				return;
			}
			System.out.println("Sign OK.");

			// 
			rc = VerifyTest(outputFile);
			if(!rc)
			{
				System.out.println("Verify Error.");
				return;
			}
			System.out.println("Verify OK.");

    	}
    	catch(Exception e)
    	{
       	    System.out.println(e);
	    	System.out.println("ERROR BY EXCEPTION!!!!!");
	    	return;
    	}

		System.out.println("Done.");
	}

    // ---------------------------------------------------------------------------
    // Ot@CDetached̎iPKCS#12t@Cgpj.
    private static boolean SignTest(
    	String target,
    	String p12file, String p12pswd,
    	String outfile) throws Exception
    {
    	// hLg̏
        DocumentBuilderFactory dbf = DocumentBuilderFactory.newInstance();
		dbf.setNamespaceAware(true);
		Document document = dbf.newDocumentBuilder().newDocument();

        // XMLSignatureFactoryDOM擾
		XMLSignatureFactory xsf = XMLSignatureFactory.getInstance("DOM");

		// Reference𐶐
		DigestMethod dm = xsf.newDigestMethod(DigestMethod.SHA256, null);
		Reference reference = xsf.newReference(target, dm, null, null, null);

		// PKCS12t@CؖƔ閧擾
		KeyStore ks = KeyStore.getInstance("PKCS12");
		FileInputStream fis = new FileInputStream(p12file);
		ks.load(fis, p12pswd.toCharArray());
		String myAlias = null;
		for (Enumeration<String> e = ks.aliases(); e.hasMoreElements() ;)
		{
			String alias = e.nextElement();
			if(ks.getKey(alias, p12pswd.toCharArray()) != null)
			{
				myAlias = alias;	// PKCS#12t@CɊ܂܂ŏɔ閧ؖalias
				break;
			}
		}
		Certificate cert = ks.getCertificate(myAlias);
		PublicKey pubKey = cert.getPublicKey();
		PrivateKey privKey = (PrivateKey)ks.getKey(myAlias, p12pswd.toCharArray());

		// KeyValueKeyInfo쐬ăZbg
		KeyInfoFactory kif = xsf.getKeyInfoFactory();
		KeyValue keyValue = kif.newKeyValue(pubKey);
		X509Data x509Data = kif.newX509Data(Collections.singletonList(cert));
		List<XMLStructure> clauses = new ArrayList<XMLStructure>();
		clauses.add((XMLStructure)keyValue);
		clauses.add((XMLStructure)x509Data);   // X.509ؖKeyInfoɊ܂߂
		KeyInfo keyInfo = kif.newKeyInfo(clauses, "MyKeyInfoId");

		// SignedInfo𐶐
		CanonicalizationMethod cm = xsf.newCanonicalizationMethod(CanonicalizationMethod.INCLUSIVE, (C14NMethodParameterSpec)null);
		SignatureMethod sm = xsf.newSignatureMethod("http://www.w3.org/2001/04/xmldsig-more#rsa-sha256", null);
        SignedInfo signedInfo = xsf.newSignedInfo(cm, sm, Collections.singletonList(reference));

		// Signaturevf쐬
		XMLSignature signature = xsf.newXMLSignature(signedInfo, keyInfo);

		// DOMpZbg
		DOMSignContext dsc = new DOMSignContext(privKey, document);

		// DetachedOt@Cł݂͌̃tH_ʒux[XURIƂ
		String path = new File(".").getCanonicalPath();
		path = "file:///" + path.replace('\\', '/') + "/";
		dsc.setBaseURI(path);

		// s
		signature.sign(dsc);

		// XMLóB
		OutputStream os = new FileOutputStream(outfile);
		TransformerFactory tf = TransformerFactory.newInstance();
		Transformer trans = tf.newTransformer();
		trans.transform(new DOMSource(document), new StreamResult(os));

		// I
    	return true;
    }

    // ---------------------------------------------------------------------------
    // Ot@CDetacheď.
	private static boolean VerifyTest(String file) throws Exception
    {
    	boolean result = false;

    	// ؃hLg̏
		DocumentBuilderFactory dbf = DocumentBuilderFactory.newInstance();
		dbf.setNamespaceAware(true);
		Document doc = dbf.newDocumentBuilder().parse(new FileInputStream(file));

		// SignaturevfT
		NodeList nl = doc.getElementsByTagNameNS(XMLSignature.XMLNS, "Signature");
		if (nl.getLength() == 0)
		{
			// SignaturevfȂ
			System.out.println("Verify: NG : cannot find Signature. : file = " + file);
            return result;
		}

		// DocumentIuWFNgAXMLSignatureIuWFNg𐶐
		XMLSignatureFactory fac = XMLSignatureFactory.getInstance("DOM");

		// ؑΏۂƌ擾NX擾iŏSignaturevfj
		Node target = nl.item(0);
		DOMValidateContext valContext = new DOMValidateContext(new KeyValueKeySelector(), target);

		// ݂̃tH_ʒux[XURIƂiDetachedOt@CŕKvƂȂj
		String path = new File(".").getCanonicalPath();
		path = "file:///" + path.replace('\\', '/') + "/";
		valContext.setBaseURI(path);

		// XML  XMLSignature 񐮗񉻂
		XMLSignature signature = fac.unmarshalXMLSignature(valContext);

		// ؎s
		boolean coreValidity = signature.validate(valContext);
		if (coreValidity == false)
		{
			// ؎s
			System.out.println("Verify: NG : faild!!!!! : file = " + file);
			// ڍ׌
			boolean sv = signature.getSignatureValue().validate(valContext);
			System.out.println(" Signature validity status: " + sv);
		}
		else
		{
			// ؐ
//			System.out.println("Verify: OK : valid. : file = " + file);
			result = true;
		}
		return result;
    }

    // ---------------------------------------------------------------------------
    // ؗp̌擾NXiłVvȎj
	public static class KeyValueKeySelector extends KeySelector
	{
		@SuppressWarnings("unchecked")
		public KeySelectorResult select(
				KeyInfo keyInfo,
				KeySelector.Purpose purpose,
				AlgorithmMethod method,
				XMLCryptoContext context) throws KeySelectorException
		{
			List<XMLStructure> list = keyInfo.getContent();

			for (int i = 0; i < list.size(); i++)
			{
				XMLStructure xmlStructure = list.get(i);
				if (xmlStructure instanceof KeyValue)
				{
					PublicKey pubKey = null;
					try
					{
						// J̎擾
						pubKey = ((KeyValue)xmlStructure).getPublicKey();
					}
					catch (KeyException ke)
					{
						throw new KeySelectorException(ke);
					}
					// ASY̊mF
					if(pubKey.getAlgorithm().equalsIgnoreCase("RSA"))
						return new SimpleKeySelectorResult(pubKey);		// OK!!
				}
			}
			throw new KeySelectorException("No KeyValue element found!");
		}
	}

    // ---------------------------------------------------------------------------
    // ؗp̌߂NXiłVvȎj
	private static class SimpleKeySelectorResult implements KeySelectorResult
	{
		private PublicKey pubKey;

		SimpleKeySelectorResult(PublicKey pubKey)
		{
			this.pubKey = pubKey;
		}

		public Key getKey()
		{
			return pubKey;
		}
	}}

